/*
******************************************************************************
**	IBM RGB52x internal clock driver
********************************************************************************
**
   Copyright	1997 David Waite
**
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
********************************************************************************
**	This code was derived from the following sources of information:
**
**		IBM's technical documentation on the IBM RGB524
**		Torgeir Veimo's beginning work on the RAMDAC driver
**		Other KGI drivers and documentation
*/

#include <kgi/maintainers.h>

#define	MAINTAINER	David_Waite
#define	DRIVER_NAME	"IBM RGB Clock driver (Old M/N)"
#define	DRIVER_REV	"$Revision: 1.10 $"

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

#include <kgi/module.h>
#include <IBM/rgb52x.h>

/*
**	These will both be autodetected soon, or perhaps specified with
**	module parameters (clock_ref, clock_sysclk)
*/ 

/*
**	Maximum/minimum frequencies
*/ 

/*defined module parameter for kgicon*/
extern kgi_uint ramdac_mhz;


#define PLL_MAX ramdac_mhz*1000000
#define PLL_MIN 16250000

kgiu32 compute_frefclk(struct kgi_display *dpy);
kgiu32 compute_fsysclk(struct kgi_display * dpy , kgiu32 frefclk);
                  
                        
#define FREFCLK private->frefclk
#define FSYSCLK private->fsysclk
#define TYPE private->type
#define GET_PRIVATE     struct _rgb52x_clock_private *private=KGIM_PRIVATE(dpy,clock)
                
                
struct 	_rgb52x_clock_private{
        kgiu32 type; 
		/* type of ramdac/clock*/
        kgiu32 frefclk; 
		/* base frequency*/
        kgiu32 fsysclk; 
		/* system function pll needed to avoid interference*/
};


static struct kgi_clock clock [4]= 	/* for 524/526 */
{
	{
		"IBM",			/* manufacturer		*/
		"RBG526-170 (old M/N)",		/* model		*/
		GM_ALL,				/* supported modes 	*/
		0,				/* special capabilities	*/
		{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		CLK_TYPE_PROG,			/* clock type		*/
		{{
			16250000, 170000000,	/* clock range */

		}}
	},
	{
		"IBM",			/* manufacturer		*/
		"RBG526-220 (old M/N)",		/* model		*/
		GM_ALL,				/* supported modes 	*/
		0,				/* special capabilities	*/
		{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		CLK_TYPE_PROG,			/* clock type		*/
		{{
		16250000, 220000000,	/* clock range */

		}}
	},
	{
		"IBM",			/* manufacturer		*/
		"RBG524-170",		/* model		*/
		GM_ALL,				/* supported modes 	*/
		0,				/* special capabilities	*/
		{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		CLK_TYPE_PROG,			/* clock type		*/
		{{
			16250000, 170000000,	/* clock range */

		}}
	},
	{
		"IBM",			/* manufacturer		*/
		"RBG524-220",		/* model		*/
		GM_ALL,				/* supported modes 	*/
		0,				/* special capabilities	*/
		{ 0x7FFF, 0x7FFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		CLK_TYPE_PROG,			/* clock type		*/
		{{
			16250000, 220000000,	/* clock range */

		}}
	},
};

/* OLD Pseudo MN fout = fref * (VCO_DIVIDE_COUNT + 65)/(REF_DIV_COUNT * DL_VAL) */
/* Rules of thumb: fref/(ref_div_count*DL_ref_VAL())>1 MHz
   the PLL clock must not be a multiple of the sysclk, or
   lower*n != higher +/- 3 Mhz */

kgi_sint kgim_clock_init(struct kgi_card *card, struct kgi_display *dpy){
 	kgiu8 id, rev;
	int type=-1;	
	struct _rgb52x_clock_private * private;
        private = kgi_malloc_small(sizeof(*private));
        if (private == NULL) {
                NOTICE2("failed to allocate private data");
                return -ENOMEM;
        }

	/*Probe */
	
	id = EXT_in8(dpy, DAC_ID);
	rev = EXT_in8(dpy, DAC_REV);
	if ((id==_RGB524_EXT01_PROD_ID)&&(rev==_RGB524_EXT00_REV_LEVEL)){
		NOTICE ("IBM 524 Integrated clock detected");
		/* set kgi_clock and init clock */
		type=2;
	}
	if ((id==_RGB524_EXT01_PROD_ID)&&(rev==_RGB524_EXT00_REV_BAD)){
		NOTICE ("IBM 524 Integrated clock detected");
		NOTICE ("INV SCLK bug detected");
		type=2;
	}
	if ((id==_RGB525_EXT01_PROD_ID)&&(rev==_RGB525_EXT00_REV_LEVEL)){
		NOTICE ("IBM 525 Integrated clock detected, this is not supported yet");
		return -E(CLOCK, UNKNOWN);
	}
	if ((id==_RGB526_EXT01_PROD_ID)&&(rev==_RGB526_EXT00_REV_LEVEL)){
		NOTICE ("IBM 526 Integrated clock detected, using old M/N method");
		type=0;
	}
	if ((id==_RGB624_EXT01_PROD_ID)&&(rev==_RGB624_EXT00_REV_LEVEL)){
		NOTICE ("IBM 624 Ramdac & Integrated clock detected, this is not supported yet");
		return -E(CLOCK, UNKNOWN);
	}
	if (type==-1){
		NOTICE("No IBM RGB CLOCK detected, or model unknown");
		DEBUG2("%i = ID, %i = REF",id,rev);
		return -E(CLOCK, UNKNOWN);
	}
	
	if (ramdac_mhz==220){
		type +=1;
		NOTICE("220 MHz Maximum clock set");
	}else {
		ramdac_mhz=170;
		NOTICE("170 MHz Maximum clock set");
	}
	card->clock=&clock[type];
        KGIM_PRIVATE(dpy,clock) = private;
        memset(private, 0, sizeof(*private));
	
	FREFCLK=compute_frefclk(dpy);
	TYPE=type;
	FSYSCLK=compute_fsysclk(dpy,FREFCLK);

	NOTICE("found %i REFCLK, %i SYSCLK",FREFCLK,FSYSCLK);

	return EOK;

}

void kgim_clock_done(struct kgi_display *dpy)
{
	/* changing this back may be done by the ramdac, since it
	 will probably be done there by a register list restore*/
        GET_PRIVATE;
	kgi_free_small(private);
        KGIM_PRIVATE(dpy, chipset) = NULL;
	 
	dpy->card->clock = NULL;
	NOTICE(DRIVER_NAME " removed");
}

/*
**	Structure for passing around clock settings
*/

struct PLL_struct{
	kgiu8 DF;
	kgiu8 vco;/*65 to 128, shifted later*/
	kgiu8 ref_div;
};

/*
**	Helper functions
*/

int inline DF_ref(int DF){ /*calculate DF contribution to fref*/
	if (DF==3) return (1);
	return (2);
};

int inline DF_vco(int DF) {/*calculate DF contribution to M/N*/
	switch (DF){
		case 0: return(4);
		case 1: return(2);
		default: return(1);
	}
};

/*for computing the fout freq resulting from the settings*/

kgiu32 compute_fout(struct PLL_struct val,kgiu32 frefclk){
	return(frefclk*val.vco /
		(val.ref_div*DF_ref(val.DF)*DF_vco(val.DF)));
};

/* check for harmonics with the exported system clock*/

int invalid_fout (kgiu32 fout,kgiu32 frefclk,kgiu32 sysclk){
	kgis32 high,low;
	kgis8 temp;
	/* fill low and high with vals*/
	if (fout<frefclk) {
		low=fout;
		high=sysclk;
	}else{
		low=sysclk;
		high=fout;
	}
	temp=(high+3000000)/low;
	/*harmonic, bad bad bad*/
	if ((temp*low-high<=3000000)&&(high-temp*low<=3000000)) return(1); 	
	/*not a harmonic, good good good*/
	return (0);
};

/*called if invalid, to find the highest frequency below the current one that is valid

note: these don't check that the value returned really is valid, 
either in the min/max sense or that they are even valid afterwards*/

kgiu32 lower_fout (kgiu32 fout,kgiu32 frefclk,kgiu32 sysclk){
	kgiu32 high,low;
	kgiu8 temp;
	/* fill low and high with vals*/
	if (fout<sysclk) {
		low=fout;
		high=sysclk;
		temp=(high+3000000)/low;
		return((high-3000000)/temp - 1);
	}else{
		low=sysclk;
		high=fout;

		temp=(high+3000000)/low;
		return (temp*low-3000001);
	}
};

kgiu32 raise_fout (kgiu32 fout,kgiu32 frefclk, kgiu32 sysclk){
	kgiu32 high,low;
	kgiu8 temp;
	/* fill low and high with vals*/
	if (fout<sysclk) {
		low=fout;
		high=sysclk;
		temp=(high+3000000)/low;
		return((high+3000000)/temp + 1);
	}else{
		low=sysclk;
		high=fout;

		temp=(high+3000000)/low;
		return (temp*low+3000001);
	}
};


/* figure out the refclk. Stolen from XFree */
kgiu32 compute_frefclk(struct kgi_display * dpy){
	int clk25_vcodf,clk25_ref;
	int clk28_vcodf,clk28_ref;
	kgiu32 f0,f1;
	kgis32 error;
	int i, rclks[] = { 14318, 16000, 24000, 28322, 50000, 0 };

	clk25_vcodf=EXT_in8(dpy,DAC_M0);
	clk25_ref=EXT_in8(dpy,DAC_N0);
	clk28_vcodf=EXT_in8(dpy,DAC_M1);
	clk28_ref=EXT_in8(dpy,DAC_N1);

	f0=25175 *clk25_ref * (8>>(clk25_vcodf>>6))/((clk25_vcodf&0x3f)+65);
	f1=28322 *clk28_ref * (8>>(clk28_vcodf>>6))/((clk28_vcodf&0x3f)+65);
	
	error=(f0-f1)*1000/f0;
	if (error>20){
		NOTICE("FREFCLK clock probe failed, defaulting to 16 MHz");
		return(16000000);	
	}
	for (i=0;i<5;i++){
		if (((rclks[i]-f1)*1000/f0)<20)
			return (rclks[i]*1000);
	}

	NOTICE("FREFCLK not matched, %i is being used",f1*1000);
	return f1*1000;
}

/*
**	Main calculating function
*/

kgiu32 compute_fsysclk(struct kgi_display *dpy,kgiu32 frefclk){
	kgiu8 sys_ref,sys_vcodf;
	sys_ref=EXT_in8(dpy,DAC_SYSCLK_REF_DIV);
	sys_vcodf=EXT_in8(dpy,DAC_SYSCLK_VCO_DIV);
	return frefclk*((sys_vcodf&0x3f)+65)/(8>>(sys_vcodf>>6))
		/sys_ref;
}
struct PLL_struct find_best_PLL (kgiu32 fout,struct kgi_display *dpy ,enum kgi_tm_cmd cmd){
	GET_PRIVATE;
	struct PLL_struct working;
	struct PLL_struct cur;
	kgiu32 fout_checked; 
	kgiu8 ref_div;
	kgis32 temp;
	kgiu32 error,temp_error;
	/*working is a temporary settings struct
	  cur is the 'current best' values
	  fout_checked is the fout value, adjusted so it doesn't
		  conflict with a harmonic
	  ref_div is the index for looking up the best PLL match in a for loop
	*/
	
	int bFlag=0,bFlag2=1;
	
	fout_checked=fout;
	if (invalid_fout(fout_checked,FREFCLK,FSYSCLK)){
		if (cmd==CMD_RAISE){
			fout_checked=raise_fout(fout,FREFCLK,FSYSCLK);
		}else 
			fout_checked=lower_fout(fout,FREFCLK,FSYSCLK);
	}
	cur.DF=cur.vco=cur.ref_div=0;
	if (invalid_fout(fout_checked,FREFCLK,FSYSCLK))
		 return cur;
	
	/*loop through ref_div from min to max*/
	for (ref_div=2;ref_div<=FREFCLK/1000000;ref_div++){
		/*instantly skip invalid vals*/
		if ((ref_div>31)&&(ref_div%2))
			ref_div++;	
		working.ref_div=ref_div;
		/*begin vco calculation*/
		temp=fout_checked/(FREFCLK/(ref_div*1000));
		/* now find vco and DF*/
		
		if ((temp>64500)&&(temp<=129000))
			working.DF=2;
		if ((temp>32250)&&(temp<=64500))
			working.DF=1;
		if ((temp>=16125)&&(temp<=32250))
			working.DF=0;
		if ((working.DF==2)&&((working.ref_div%2)==1))
			working.DF=3;
		if ((temp<16125)||(temp>129000)) continue;
		bFlag=0;
		switch(working.DF){
			case 0: if (fout_checked>(PLL_MAX/4)) bFlag=1;
				break;
			case 1: if (fout_checked>(PLL_MAX/2)) bFlag=1;
				break;
			case 2:
			case 3: if (fout_checked>(PLL_MAX)) bFlag=1;
				break;
		}
		if (bFlag==1) {
			continue;
		}
		working.vco=(kgiu8)(temp*DF_vco(working.DF)/1000);
		working.ref_div/=DF_ref(working.DF);
		/* now compute the PLL-out resulting */
		if (working.vco>64){
			temp=compute_fout(working,FREFCLK);
			temp_error=(temp-fout_checked);
			if(((cmd==CMD_RAISE)&&(temp-(signed)fout_checked>0))||
				((cmd==CMD_LOWER)&&((signed)fout_checked-temp>0))||
				((cmd!=CMD_LOWER)&&(cmd!=CMD_RAISE))){
				if (temp_error<0) temp_error=-temp_error;
				if ((temp_error<error)||(bFlag2)){
					error=temp_error;
					bFlag2=0;
					cur=working;
				}
			}
		}
		working.vco++;
		if (working.vco<129){
			temp=compute_fout(working,FREFCLK);
			temp_error=(temp-fout_checked);
			if(((cmd==CMD_RAISE)&&(temp-(signed)fout_checked>0))||
				((cmd==CMD_LOWER)&&((signed)fout_checked-temp>0))||
				((cmd!=CMD_RAISE)&&(cmd!=CMD_LOWER))){
				if (temp_error<0) temp_error=-temp_error;
				if ((temp_error<error)||(bFlag2)){
					error=temp_error;
					bFlag2=0;
					cur=working;
				}
			}
		}
	}
	return (cur);
};

static kgiu16 pll_command(struct PLL_struct s)
{
	/*need.. */
	/*two bytes for DF */
	/*5 for ref_div */
	/*6 for vco_div */
	/*=13 */
		ASSERT (s.DF<=3);
	ASSERT (s.vco>=65);
	ASSERT (s.vco<=128);
	ASSERT (s.ref_div<=31);
	DEBUG2("PLL_COMMAND: DF= %i ,vco(pre)= %i, ref= %i",
		s.DF,s.vco,s.ref_div);
	/* format:
	DF		=1
	vco_div	=2
	ref_div	=3
	|1|1|2|2| |2|2|2|2| |0|0|0|3| |3|3|3|3|
	*/
	return ( 
	  ((s.DF)<<14)
	| ((s.vco-65)<<8)
	| (s.ref_div)
	); 
};

kgi_sint kgim_clock_command (struct kgi_graphic *state, kgi_uint cmd, kgi_uint arg){
	return -E(CLOCK,UNKNOWN);
};

void PLL_set(struct kgi_display * dpy , kgiu16 clock){
	kgiu8 vcoDF=(clock&0xff00)>>8;
	kgiu8 ref=(clock&0x00ff);
	
	DEBUG3("PLL_set, cmd = %i, vcodf= %i, ref= %i",clock,vcoDF,ref);
	/* set to 8 clock mode */
	EXT_out8(dpy,_RGB52X_EXT10_PLL_INT_FS_M_N,DAC_PLL_CTRL1);
	/* check to see if the clock is in the correct mode, if not, change it*/

	EXT_out8(dpy,
		vcoDF,DAC_M2);
	EXT_out8(dpy,ref,DAC_N2);
	EXT_out8(dpy,_RGB52X_EXT11_PLL_INT_FS(2),DAC_PLL_CTRL2);
}

void kgim_clock_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
	DEBUG2("KGIM_CLOCK_SET_MODE");
	if (mode->clock!=0)
		PLL_set(dpy, mode->clock);
}	

/* taken from pll.inc*/
kgi_sint kgim_clock_check_mode(struct kgi_display *dpy, struct kgi_mode *mode, 
	enum kgi_tm_cmd *cmd)
{
	GET_PRIVATE;
	struct PLL_struct state;

	kgiu32 newfreq = (*cmd == CMD_PROPOSE) 
		? 0 : mode->dclk * mode->clk.mul / mode->clk.div;
	switch (*cmd) {

	case CMD_PROPOSE:
		DEBUG2("CMD_PROPOSE");
		mode->clock = 0;
		return EOK;

	case CMD_LOWER:
		DEBUG2("CMD_LOWER");
		if (newfreq < PLL_MIN) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too low",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq > PLL_MAX) {

			newfreq = PLL_MAX;
		}

		state=find_best_PLL(newfreq,dpy, *cmd);
		if (state.ref_div==0){
			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = compute_fout(state,FREFCLK) * mode->clk.div / mode->clk.mul;
		return EOK;

	case CMD_RAISE:
		DEBUG2("CMD_RAISE");
		if (PLL_MAX < newfreq) {

			DEBUG1("DCLK = %i Hz, CLK = %i Hz is too high", \
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		if (newfreq < PLL_MIN) {

			newfreq = PLL_MIN;
		}

		state=find_best_PLL(newfreq,dpy, *cmd);
		if (state.ref_div==0){
			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->dclk = compute_fout(state,FREFCLK) * mode->clk.div / mode->clk.mul;

		return EOK;
		
	case CMD_CHECK:
		DEBUG2("CMD_CHECK");
		mode->clock = 0;
		state=(find_best_PLL(newfreq,dpy,*cmd));

		if (state.ref_div==0){
			DEBUG1("invalid request for DCLK = %i Hz, CLK = %i Hz",\
				mode->dclk, newfreq);
			return -E(CLOCK, UNKNOWN);
		}

		mode->clock = pll_command(state);
		return EOK;

	default:
		DEBUG2("CMD_UNKNOWN");
		INTERNAL_ERROR;
		return -E(CLOCK, UNKNOWN);
	}
};
